@@ -18,6 +18,7 @@
 /* exported DesktopManager */
 'use strict';
 const GLib = imports.gi.GLib;
+const GLibUnix = imports.gi.GLibUnix;
 const Gtk = imports.gi.Gtk;
 const Gdk = imports.gi.Gdk;
 const Gio = imports.gi.Gio;
@@ -37,6 +38,7 @@ const TemplatesScriptsManager = imports.templatesScriptsManager;
 const Thumbnails = imports.thumbnails;
 const FileItemMenu = imports.fileItemMenu;
 const AutoAr = imports.autoAr;
+const SignalManager = imports.signalManager;
 
 const Gettext = imports.gettext.domain('ding');
 
@@ -45,6 +47,7 @@ const _ = Gettext.gettext;
 var DesktopManager = class {
     constructor(mainApp, dbusManager, desktopList, codePath, asDesktop, primaryIndex) {
         this.mainApp = mainApp;
+        this._lastSelected = null;
         this.using_X11 = Gdk.Display.get_default().constructor.$gtype.name === 'GdkX11Display';
         if (asDesktop) {
             this.mainApp.hold(); // Don't close the application if there are no desktops
@@ -261,7 +264,7 @@ var DesktopManager = class {
         }
         this._pendingDropFiles = {};
         if (this._asDesktop) {
-            this._sigtermID = GLib.unix_signal_add(GLib.PRIORITY_DEFAULT, 15, () => {
+            this._sigtermID = GLibUnix.signal_add_full(GLib.PRIORITY_DEFAULT, 15, () => {
                 GLib.source_remove(this._sigtermID);
                 for (let desktop of this._desktops) {
                     desktop.destroy();
@@ -394,9 +397,10 @@ var DesktopManager = class {
             let desktop = this._desktopList[desktopIndex];
             let desktopName;
             if (this._asDesktop) {
-                desktopName = `@!${desktop.x},${desktop.y};BDHF`;
+                // this name must match the one used in emulateX11WindowType
+                desktopName = `Desktop Icons ${desktop.monitorIndex + 1}`;
             } else {
-                desktopName = `DING ${desktopIndex}`;
+                desktopName = `DING ${desktop.monitorIndex + 1}`;
             }
             this._desktops.push(new DesktopGrid.DesktopGrid(this, desktopName, desktop, this._asDesktop, this._premultiplied));
         }
@@ -791,36 +795,92 @@ var DesktopManager = class {
             console.log("Mismatched hidePopup() and showPopup() calls");
     }
 
+    _getTopLeftIcon() {
+        if (this._fileList.length == 0) {
+            return null;
+        }
+        let currentCoords = null;
+        let currentItem = null;
+        for (let item of this._fileList) {
+            const newCoords = item.getCoordinates();
+            if ((currentCoords === null) || (newCoords[0] < currentCoords[0]) || (newCoords[1] < currentCoords[1])) {
+                currentCoords = newCoords;
+                currentItem = item;
+            }
+        }
+        return currentItem;
+    }
+
+    _getBottomRightIcon() {
+        if (this._fileList.length == 0) {
+            return null;
+        }
+        let currentCoords = null;
+        let currentItem = null;
+        for (let item of this._fileList) {
+            const newCoords = item.getCoordinates();
+            if ((currentCoords === null) || (newCoords[0] > currentCoords[0]) || (newCoords[1] > currentCoords[1])) {
+                currentCoords = newCoords;
+                currentItem = item;
+            }
+        }
+        return currentItem;
+    }
+
+    _setIconAsSelected(icon) {
+        this._fileList.forEach(fileItem => fileItem.isKeyboardSelected = fileItem === icon);
+    }
+
+    _getLastKeyboardIcon() {
+        if ((this._lastSelected !== null) && this._fileList.includes(this._lastSelected)) {
+            this._setIconAsSelected(this._lastSelected);
+            return this._lastSelected;
+        }
+        return null;
+    }
+
+    _getCurrentKeyboardIcon() {
+        let currentKeyboardIcon = null;
+
+        for (let fileItem of this._fileList) {
+            if ((currentKeyboardIcon === null) && (fileItem.isKeyboardSelected)) {
+                currentKeyboardIcon = fileItem;
+            } else {
+                if (fileItem.isKeyboardSelected) {
+                    fileItem.isKeyboardSelected = false;
+                }
+            }
+        }
+        return currentKeyboardIcon;
+    }
+
     onKeyRelease(event, grid) {
         if (this._popupCounter != 0)
             return false;
-        let symbol = event.get_keyval()[1];
-        let selection = this.getCurrentSelection(false);
+        const isCtrl = (event.get_state()[1] & Gdk.ModifierType.CONTROL_MASK) != 0;
+        const isShift = (event.get_state()[1] & Gdk.ModifierType.SHIFT_MASK) != 0;
+        const symbol = event.get_keyval()[1];
         if ((symbol == Gdk.KEY_Left) || (symbol == Gdk.KEY_Right) ||
-        (symbol == Gdk.KEY_Up) || (symbol == Gdk.KEY_Down)) {
-            if (!selection) {
-                selection = this._fileList;
-            }
-            if (!selection) {
-                return false;
-            }
-            let selected = selection[0];
-            let selectedCoordinates = selected.getCoordinates();
-            this.unselectAll();
-            if (selection.length > 1) {
-                for (let item of selection) {
-                    let itemCoordinates = item.getCoordinates();
-                    if (itemCoordinates[0] > selectedCoordinates[0]) {
-                        continue;
-                    }
-                    if ((itemCoordinates[0] < selectedCoordinates[0]) ||
-                        (itemCoordinates[1] < selectedCoordinates[1])) {
-                        selected = item;
-                        selectedCoordinates = itemCoordinates;
-                        continue;
-                    }
+           (symbol == Gdk.KEY_Up) || (symbol == Gdk.KEY_Down)) {
+            let selected = this._getCurrentKeyboardIcon();
+            // if there is no selected icon, select the last selected icon
+            if (!selected) {
+                selected = this._getLastKeyboardIcon();
+                if (selected) {
+                    return false;
                 }
             }
+            // if there is no last selected, or the last selected isn't in the desktop
+            // (for example, because it was deleted), select the top-left icon.
+            if (!selected) {
+                selected = this._getTopLeftIcon();
+                if (selected) {
+                    selected.isKeyboardSelected = true;
+                }
+                this._lastSelected = selected;
+                return false;
+            }
+            let selectedCoordinates = selected.getCoordinates();
             let index;
             let multiplier;
             switch (symbol) {
@@ -856,22 +916,42 @@ var DesktopManager = class {
             }
             if (newItem === null) {
                 newItem = selected;
+            } else {
+                selected.isKeyboardSelected = false;
+                if (isCtrl || isShift) {
+                    selected.setSelected();
+                }
             }
-            newItem.setSelected();
+            newItem.isKeyboardSelected = true;
+            this._lastSelected = newItem;
             return false;
         }
         return false;
     }
 
     onKeyPress(event, grid) {
-        if (this._popupCounter != 0)
+        if (this._popupCounter != 0) {
             return false;
-        let symbol = event.get_keyval()[1];
-        let isCtrl = (event.get_state()[1] & Gdk.ModifierType.CONTROL_MASK) != 0;
-        let isShift = (event.get_state()[1] & Gdk.ModifierType.SHIFT_MASK) != 0;
-        let isAlt = (event.get_state()[1] & Gdk.ModifierType.MOD1_MASK) != 0;
-        let selection = this.getCurrentSelection(false);
-        if (isCtrl && isShift && ((symbol == Gdk.KEY_Z) || (symbol == Gdk.KEY_z))) {
+        }
+
+        const symbol = event.get_keyval()[1];
+        const isCtrl = (event.get_state()[1] & Gdk.ModifierType.CONTROL_MASK) != 0;
+        const isShift = (event.get_state()[1] & Gdk.ModifierType.SHIFT_MASK) != 0;
+        const isAlt = (event.get_state()[1] & Gdk.ModifierType.MOD1_MASK) != 0;
+        const selection = this.getCurrentSelection(false);
+        if (symbol == Gdk.KEY_Home) {
+            this._setIconAsSelected(this._getTopLeftIcon());
+            return true;
+        } else if (symbol == Gdk.KEY_End) {
+            this._setIconAsSelected(this._getBottomRightIcon());
+            return true;
+        } else if (isCtrl && (symbol === Gdk.KEY_space)) {
+            const selected = this._getCurrentKeyboardIcon();
+            if (selected !== null) {
+                selected.toggleSelected();
+                return true;
+            }
+        } else if (isCtrl && isShift && ((symbol == Gdk.KEY_Z) || (symbol == Gdk.KEY_z))) {
             this._doRedo();
             return true;
         } else if (isCtrl && ((symbol == Gdk.KEY_Z) || (symbol == Gdk.KEY_z))) {
@@ -936,7 +1016,7 @@ var DesktopManager = class {
         } else if (isCtrl && isShift && ((symbol == Gdk.KEY_N) || (symbol == Gdk.KEY_n))) {
             this.doNewFolder();
             return true;
-        } else if (symbol == Gdk.KEY_Menu) {
+        } else if ((symbol == Gdk.KEY_Menu) || ((symbol == Gdk.KEY_F10) && (isShift))) {
             if (selection) {
                 this.fileItemMenu.showMenu(selection[0], event, true);
             } else {
@@ -988,7 +1068,10 @@ var DesktopManager = class {
     }
 
     unselectAll() {
-        this._fileList.map(f => f.unsetSelected());
+        this._fileList.map(f => {
+            f.unsetSelected();
+            f.isKeyboardSelected = false;
+        });
     }
 
     findFiles(text) {
@@ -1007,12 +1090,13 @@ var DesktopManager = class {
         this._findFileTextArea = new Gtk.Entry();
         contentArea.pack_start(this._findFileTextArea, true, true, 5);
         contentArea = undefined;
-        this._findFileTextArea.connect('activate', () => {
+        this._findFileSignalManager = new SignalManager.SignalManager();
+        this._findFileSignalManager.connectSignal(this._findFileTextArea, 'activate', () => {
             if (this._findFileButton.sensitive) {
                 this._findFileWindow.response(Gtk.ResponseType.OK);
             }
         });
-        this._findFileTextArea.connect('changed', () => {
+        this._findFileSignalManager.connectSignal(this._findFileTextArea, 'changed', () => {
             let context = this._findFileTextArea.get_style_context();
             if (this.scanForFiles(this._findFileTextArea.text, true)) {
                 this._findFileButton.sensitive = true;
@@ -1036,13 +1120,14 @@ var DesktopManager = class {
             this.scanForFiles(null);
         }
         this._findFileWindow.show_all();
-        this._findFileWindow.connect('close', () => {
+        this._findFileSignalManager.connectSignal(this._findFileWindow, 'close', () => {
             this._findFileWindow.response(Gtk.ResponseType.CANCEL);
         });
-        this._findFileWindow.connect('response', (actor, retval) => {
+        this._findFileSignalManager.connectSignal(this._findFileWindow, 'response', (actor, retval) => {
             if (retval == Gtk.ResponseType.CANCEL) {
                 this.unselectAll();
             }
+            this._findFileSignalManager.disconnectAllSignals();
             this._findFileWindow.destroy();
             this._findFileWindow = null;
         });
@@ -1110,7 +1195,7 @@ var DesktopManager = class {
 
         this._changeBackgroundMenuItem = new Gtk.MenuItem({label: _('Change Background…')});
         this._changeBackgroundMenuItem.connect('activate', () => {
-            let desktopFile = Gio.DesktopAppInfo.new('gnome-background-panel.desktop');
+            const desktopFile = Gio.DesktopAppInfo.new('gnome-background-panel.desktop');
             const context = Gdk.Display.get_default().get_app_launch_context();
             context.set_timestamp(Gtk.get_current_event_time());
             desktopFile.launch([], context);
@@ -1120,12 +1205,21 @@ var DesktopManager = class {
         this._menu.add(new Gtk.SeparatorMenuItem());
 
         this._settingsMenuItem = new Gtk.MenuItem({label: _('Desktop Icons Settings')});
-        this._settingsMenuItem.connect('activate', () => Prefs.showPreferences());
+        if (GLib.getenv('XDG_CURRENT_DESKTOP').split(':').includes('ubuntu')) {
+            this._settingsMenuItem.connect("activate", () => {
+                const desktopFile = Gio.DesktopAppInfo.new('gnome-ubuntu-panel.desktop');
+                const context = Gdk.Display.get_default().get_app_launch_context();
+                context.set_timestamp(Gtk.get_current_event_time());
+                desktopFile.launch([], context);
+            });
+        } else {
+            this._settingsMenuItem.connect("activate", () => Prefs.showPreferences());
+        }
         this._menu.add(this._settingsMenuItem);
 
         this._displaySettingsMenuItem = new Gtk.MenuItem({label: _('Display Settings')});
         this._displaySettingsMenuItem.connect('activate', () => {
-            let desktopFile = Gio.DesktopAppInfo.new('gnome-display-panel.desktop');
+            const desktopFile = Gio.DesktopAppInfo.new('gnome-display-panel.desktop');
             const context = Gdk.Display.get_default().get_app_launch_context();
             context.set_timestamp(Gtk.get_current_event_time());
             desktopFile.launch([], context);
@@ -1646,7 +1740,7 @@ var DesktopManager = class {
     }
 
     doTrash() {
-        const selection = this._fileList.filter(i => i.isSelected && !i.isSpecial).map(i =>
+        const selection = this._fileList.filter(i => (i.isSelected || i.isKeyboardSelected) && !i.isSpecial).map(i =>
             i.file.get_uri());
 
         if (selection.length) {
@@ -1655,11 +1749,11 @@ var DesktopManager = class {
     }
 
     doDeletePermanently() {
-        const toDelete = this._fileList.filter(i => i.isSelected && !i.isSpecial).map(i =>
+        const toDelete = this._fileList.filter(i => (i.isSelected || i.isKeyboardSelected) && !i.isSpecial).map(i =>
             i.file.get_uri());
 
         if (!toDelete.length) {
-            if (this._fileList.some(i => i.isSelected && i.isTrash)) {
+            if (this._fileList.some(i => (i.isSelected || i.isKeyboardSelected) && i.isTrash)) {
                 this.doEmptyTrash();
             }
             return;
@@ -1683,7 +1777,7 @@ var DesktopManager = class {
 
     checkIfDirectoryIsSelected() {
         for (let item of this._fileList) {
-            if (item.isSelected && item.isDirectory) {
+            if ((item.isSelected || item.isKeyboardSelected) && item.isDirectory) {
                 return true;
             }
         }
@@ -1693,7 +1787,7 @@ var DesktopManager = class {
     getCurrentSelection(getUri) {
         let listToTrash = [];
         for (let fileItem of this._fileList) {
-            if (fileItem.isSelected) {
+            if ((fileItem.isSelected) || (fileItem.isKeyboardSelected)) {
                 if (getUri) {
                     listToTrash.push(fileItem.file.get_uri());
                 } else {
@@ -1711,7 +1805,7 @@ var DesktopManager = class {
     getNumberOfSelectedItems() {
         let count = 0;
         for (let item of this._fileList) {
-            if (item.isSelected) {
+            if ((item.isSelected) || (item.isKeyboardSelected)) {
                 count++;
             }
         }
